;-------------------------------------------------------
;	Iron Spring PL/I Runtime Library Version 0.5
;	   Distributed under the Gnu LGPL License
;
;	_pli_Math - PL/I runtime support routines
;       Version 0.8A Beta, 15 Dec, 2008
;       Copyright Peter Flass
;
;	This procedure is called to perform various math functions for 
;	which the compiler does not generate the appropriate instructions.
;
;	Calling Sequence:
;	  DCL MATH entry( fixed bin(31), float bin(64) )
;	           returns( float bin(64)
;	           ext( '_pli_Math')
;	           options(linkage(system));
;	  The first argument is the function to be performed 
;	     (see jmptab).
;	  The second argument is the function argument.
;
;	To Do:
;	
;       Modifications:
;	
;-------------------------------------------------------
	.486P
	.model flat,syscall
        public _pli_Math
	.data
_pli_data equ $
	include framedef.inc
;------------------------------------------------+
;  Unique Stack Data for _pli_Math               |
;------------------------------------------------+
frm_loc equ     frm_bos		; Start of locals for _pli_Math
loc_fpc equ     frm_loc-4	; Saved fpucw
loc_end equ     frm_loc-4 	; end of local stack
loc_len equ     frm_loc-loc_end	; Length of local stack

	.code			; PL/I Compatible entry code
_pli_code equ $
	byte 20h,08h,12h,15h,09h,30h,00h,00h
; 
;-----------------------------------------------------
;	_pli_Math:  PL/I Math Functions
;-----------------------------------------------------
	byte   '_pli_Math'		; Entry point name
	byte   9			; Length of name
_pli_Math:
	call dword ptr 0[edi]		; Init stack frame
	dd loc_len			; DSA size
	mov word ptr -28[ebp],81E1h	; Condition prefix flags
	fstcw loc_fpc[ebp]		; Save fpcw
	fldcw fpmask			; Set new value
	fld   tbyte ptr 12[ebp]		; Load argument
	mov   eax,8[ebp]		; Load function code
;					  (codes are arbitrary)
	jmp  jmptab[eax*4]		; Go perform function
jmptab	equ  $
	dd   offset acos		;  0=ACOS
	dd   offset asin		;  1=ASIN
	dd   offset atan		;  2=ATAN
	dd   offset atand		;  3=ATAND
	dd   offset atanh		;  4=ATANH
	dd   offset cos  		;  5=COS  
	dd   offset cosd 		;  6=COSD 
	dd   offset cosh 		;  7=COSH 
	dd   offset erf  		;  8=ERF  
	dd   offset erfc 		;  9=ERFC 
	dd   offset exp  		; 10=EXP  
	dd   offset log  		; 11=LOG  
	dd   offset log2 		; 12=LOG2 
	dd   offset log10		; 13=LOG10
	dd   offset sin  		; 14=SIN 
	dd   offset sind 		; 15=SIND 
	dd   offset sinh 		; 16=SINH 
	dd   offset sqrt 		; 17=SQRT 
	dd   offset tan  		; 18=TAN  
	dd   offset tand 		; 19=TAND
	dd   offset tanh 		; 20=TANH
	dd   offset gamma		; 21=GAMMA
	dd   offset loggamma		; 22=LOGGAMMA
; 
;-----------------------------------------------------
;	Functional Routines
;	Not implemented: acos, asin, atan, atand, atanh,
;	  cosh (need e), erf, erfc, exp,
;	  sinh (need e), tanh (need e), gamma, loggama.            
;-----------------------------------------------------
cos: 	fcos				; cos(x) in radians
	jmp short  retrn
cosd:	fcos				; cos(x) in radians
	jmp short  deg			; convert to degrees and exit
;	Compute loge(x).  This takes advantage of the identity:
;	logb(x) = logk(x)/logk(b), where b is the desired log base
;	and k is any arbitrary log base.  In this case loge(x)=log2(x)/log2(e)
log:	fld1				; Load 1.0 for fyl2x instruction
	fxch				; ST(1)<-1.0, ST(0)<-x
	fyl2x				; ST(0)<-log2(x)
	fldl2e				; ST(0)<-log2(e), ST(1)<-log2(x)
	fdiv				; ST(0)<-loge(x)
	jmp short  retrn
log2:	fld1				; Load 1.0
	fxch				; ST(1)<-1.0, ST(0)<-x
	fyl2x				; ST(0)<-log2(x)
	jmp short  retrn
;	Compute log10(x), see note and comments for log function.
log10:	fld1
	fxch
	fyl2x
	fldl2t				; ST(0)<-log2(10), ST(1)<-log2(x)
	fdiv				; ST(0)<-log10(x)
	jmp short  retrn
sin:	fsin				; sine(x) in radians
	jmp short  retrn
sind:	fsin				; sine(x) in radians
	jmp short  deg			; Convert to degrees and exit
sqrt:	fsqrt				; ST(0)<-sqrt(x)
	jmp short  retrn
tan:	fptan				; Partial tan(x) in ST(1), 1.0 in ST(0)
	fstp st(0)			; Pop the FPU stack
	jmp short  retrn
tand:	fptan				; Partial tan(x) in ST(1), 1.0 in ST(0)
	fstp st(0)			; Pop the FPU stack
	jmp short  deg			; Convert to degrees and exit
; 
;-----------------------------------------------------
;	Unimplemented Functions
;-----------------------------------------------------
acos:
asin:
atan:
atand:
atanh:
cosh:
erf:
erfc:
exp:
sinh:
tanh:
gamma:
loggamma:
	jmp short  retrn		; (Just exit for now)
	
; 
;-----------------------------------------------------
;	Return to caller               
;-----------------------------------------------------
deg  	equ  $				; Convert value in ST(0) to degrees
	fld  tbyte ptr pi_180		; ST(0)<-2pi/360
	fdiv				; Convert radians to degrees
;					; Fall-thru to return	
retrn	equ  $
	fldcw loc_fpc[ebp]		; Restore fpcw
	mov ebx,dword ptr -12[ebp]
	mov esi,dword ptr -8[ebp]  
	mov edi,dword ptr -4[ebp]  
	leave
	ret

	.data
	align	2
fpmask  db      20h,0Bh				; FPU Control word
; 	LOB (exception masks):
; 	..1. .... PE Precision mask   '20'bx
; 	...1 .... UE Underflow mask   '10'bx
; 	.... 1... OE Overflow mask    '08'bx
; 	.... .1.. ZE Zerodivide mask  '04'bx
; 	.... ..1. DE Denormal Op mask '02'bx
; 	.... ...1 IE Invalid Op mask  '01'bx
; 	HOB (control flags):
; 	...1 .... Infinity Control '01'bx
; 	.... xx.. Rounding Control
;		'11'bx - Round toward zero (chop mode)
; 	.... ..xx Precision Control 
;		'11'bx - Double Extended Precision
;						  Double Extended Precision
;						  Round toward nearest
;						  #P masked

;	Hexadecimal value of 2*PI/360 [=PI/180] (1.7453292520E-0002)
;	to convert radians to degrees
pi_180	byte	0ADh,0C8h,0E9h,094h,012h,035h,0FAh,08Eh,0F9h,03Fh

	.code
_pli_endc equ $
	.data
	org _pli_data+00h
_pli_endd equ $
	end


